package edu.toronto.cs;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import edu.toronto.cs.parsers.ProofPrunerParser;
import edu.toronto.cs.parsers.ProofReindexParser;

import jargs.gnu.CmdLineParser;

public class ProofPruner {

	public static DefaultHandler handler;
	public static ArrayList<String> rootClauses;
	public static HashMap<String, Integer> rootClauseLabels;
	public static HashMap<Integer, Integer> clauseChildren;

	public static SAXParserFactory factory = SAXParserFactory.newInstance();

	private static void printUsage() {
		System.err.println("Usage: ProofPruner [{-r,--proof} proof_file] [{-o,--output} pruned_proof]\n"
				+ "[{--periph}] [{-m,--model} model_file] [{-p,--property} property_file]\n"
				+ "[{--reindex}]");
	}

	public static void main(String[] args) {
		CmdLineParser parser = new CmdLineParser();

		// inputs
		CmdLineParser.Option optProof = parser.addStringOption('r', "proof");
		CmdLineParser.Option optOutput = parser.addStringOption('o', "output");

		CmdLineParser.Option optPeriph = parser.addBooleanOption("periph");
		CmdLineParser.Option optModel = parser.addStringOption('m', "model");
		CmdLineParser.Option optProperty = parser.addStringOption('p', "property");

		CmdLineParser.Option optReindex = parser.addBooleanOption("reindex");

		try {
			parser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(2);
		}

		String filenameProof = (String) parser.getOptionValue(optProof);
		String filenameOutput = (String) parser.getOptionValue(optOutput);

		Boolean periph = (Boolean) parser.getOptionValue(optPeriph, Boolean.FALSE);
		String filenameModel = (String) parser.getOptionValue(optModel);
		String filenameProperty = (String) parser.getOptionValue(optProperty);

		Boolean reindex = (Boolean) parser.getOptionValue(optReindex, Boolean.FALSE);

		// for normal and periph options, this proof has previously been inverted using helperScripts/invertProof.pl
		// for reindex option, proof must be in regular order
		File proofFile = new File(filenameProof);

		if (!reindex) {
			try {
				rootClauses = new ArrayList<String>(10000);
				// pruned proof is stored in filenameOutput as the original
				// proof is parsed.
				clauseChildren = new HashMap<Integer, Integer>();
				handler = new ProofPrunerParser(filenameOutput, false);

				SAXParser saxParser = factory.newSAXParser();
				saxParser.parse(proofFile, handler);

				if (periph) {
					rootClauseLabels = new HashMap<String, Integer>();
					int size = rootClauses.size();
					for (int i = 0; i < size; i++) {
						rootClauseLabels.put(rootClauses.get(i), (Integer) 0);
					}

					File modelFile = new File(filenameModel);
					File propertyFile = new File(filenameProperty);

					// types: Model = 1, Property = 2
					// thus, clauses present in both files will be labeled 3
					getClauseLabels(modelFile, 1);
					getClauseLabels(propertyFile, 2);

					// pruned proof is stored in filenameOutput, adding model
					// and property info.
					handler = new ProofPrunerParser(filenameOutput, true);

					saxParser = factory.newSAXParser();
					saxParser.parse(proofFile, handler);
				}

			} catch (SAXException ex) {
				System.err.println("SAXException " + ex.getMessage());
				System.exit(2);
			} catch (ParserConfigurationException ex) {
				System.err.println("ParserCopnfigurationException " + ex.getMessage());
				System.exit(2);
			} catch (IOException ex) {
				System.err.println("IOException " + ex.getMessage());
				System.exit(2);
			}
		}
		else{
			try {
				// proof with new indexes is stored in filenameOutput as the original
				// proof is parsed.
				handler = new ProofReindexParser(filenameOutput);

				SAXParser saxParser = factory.newSAXParser();
				saxParser.parse(proofFile, handler);
			} catch (SAXException ex) {
				System.err.println("SAXException " + ex.getMessage());
				System.exit(2);
			} catch (ParserConfigurationException ex) {
				System.err.println("ParserCopnfigurationException " + ex.getMessage());
				System.exit(2);
			} catch (IOException ex) {
				System.err.println("IOException " + ex.getMessage());
				System.exit(2);
			}			
		}
	}

	private static void getClauseLabels(File file, int type) {
		String line;
		String c;
		try {
			BufferedReader in = new BufferedReader(new FileReader(file));

			while ((line = in.readLine()) != null) {
				
				if (line.startsWith("c") || line.startsWith("p"))
					continue;

				String[] args = line.split(" ");

				// CNF files are in SMV format
				// Will turn SMV indexes into MiniSat indexes before doing
				// anything else
				// (SMV_ind = MiniSat_ind + 1)
				// -43 --> -42, 43 --> 42
				ArrayList<Integer> clause = new ArrayList<Integer>();
				int index;
				for (int i = 0; i < args.length - 1; i++) {
					index = Integer.parseInt(args[i]);
					if (index < 0)
						index++;
					else 
						index--;

					clause.add(index);

				}

				// now we test if the clause is actually in the UNSAT core.
				// If so, we keep its initial label
				// Model = 1, Property = 2, Both = 3 
				Collections.sort(clause);
				c = clause.toString();
				if (rootClauses.contains(c))
					rootClauseLabels.put(c, rootClauseLabels.get(c) + type);

			}

			in.close();

		} catch (Exception ex) {
		}

	}

}
