package edu.toronto.cs;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;

import edu.toronto.cs.algorithms.Irrelevance;
import edu.toronto.cs.algorithms.LocalIrrelevance;
import edu.toronto.cs.algorithms.Peripherality;
import edu.toronto.cs.info.Variable;

import jargs.gnu.CmdLineParser;

public class VaqTree {

	private static List<Variable> results;

	private static void printUsage() {
		System.err.println("Usage: VaqTree [{-m,--model} model_file] [{-p,--property} property_file] "
				+ "[{-r,--proof} proof_file]\n [{-v,--variables} variable_file] [{-d,--outputDirectory} outputDirectory] "
				+ "[{-s,--outputSuffix} suffix]\n [{--irrel}] [{--localIrrel}] [{--periph}] [{--nmrP_i x}] [{--lastIndex x}]\n\n"
				+ "Analysis:\nresults stored in <outputDirectory>/output_<analysis>_<suffix>\n" + "\n--irrel\tturns on irrelevance analysis"
				+ "\n--localIrrel\tturns on local irrelevance analysis" + "\n--periph\tturns on peripherality analysis\n\n"
				+ "If no input files are specified, default filenames in <outputDirectory> will be used:\n"
				+ "--model model.dimacs\n--property property.dimacs\n--proof proof.xml\n--variables variables.txt\n");
	}

	public static void main(String[] args) throws Exception {

		CmdLineParser parser = new CmdLineParser();

		// inputs
		CmdLineParser.Option optModel = parser.addStringOption('m', "model");
		CmdLineParser.Option optProperty = parser.addStringOption('p', "property");
		CmdLineParser.Option optProof = parser.addStringOption('r', "proof");
		CmdLineParser.Option optVariables = parser.addStringOption('v', "variables");
		CmdLineParser.Option optDirectory = parser.addStringOption('d', "outputDirectory");
		CmdLineParser.Option optSuffix = parser.addStringOption('s', "outputSuffix");

		// possible algorithms
		CmdLineParser.Option optIrrel = parser.addBooleanOption("irrel");
		CmdLineParser.Option optLocalIrrel = parser.addBooleanOption("localIrrel");
		CmdLineParser.Option optPeriph = parser.addBooleanOption("periph");
		CmdLineParser.Option optPeriphNmrP_i = parser.addIntegerOption("nmrP_i");
		CmdLineParser.Option optPeriphLastIndex = parser.addIntegerOption("lastIndex");

		try {
			parser.parse(args);
		} catch (CmdLineParser.OptionException e) {
			System.err.println(e.getMessage());
			printUsage();
			System.exit(2);
		}

		// default: no analysis will be carried out
		Boolean irrel = (Boolean) parser.getOptionValue(optIrrel, Boolean.FALSE);
		Boolean localIrrel = (Boolean) parser.getOptionValue(optLocalIrrel, Boolean.FALSE);
		Boolean periph = (Boolean) parser.getOptionValue(optPeriph, Boolean.FALSE);
		Integer nmrP_i = (Integer) parser.getOptionValue(optPeriphNmrP_i, 1);
		Integer lastIndex = (Integer) parser.getOptionValue(optPeriphLastIndex);

		// default: if no files are specified, will use default names
		// (model.dimacs, property.dimacs, proof.xml, variables.txt) in
		// outputDirectory
		String directory = (String) parser.getOptionValue(optDirectory);
		String filenameModel = (String) parser.getOptionValue(optModel, directory + "/model.dimacs");
		String filenameProperty = (String) parser.getOptionValue(optProperty, directory + "/property.dimacs");
		String filenameProof = (String) parser.getOptionValue(optProof, directory + "/proof.xml");
		String filenameVariables = (String) parser.getOptionValue(optVariables, directory + "/variables.txt");

		String suffix = (String) parser.getOptionValue(optSuffix);

		File modelFile = new File(filenameModel);
		File propertyFile = new File(filenameProperty);
		File proofFile = new File(filenameProof);
		File variableFile = new File(filenameVariables);

		// get variables that will be used in the analysis
		// NB: variables.txt has SMV indexes (SMV_ind = MiniSat_ind + 1)
		// Will turn SMV indexes into MiniSat indexes before doing anything else 

		// file format:
		//    model variable name
		//    p_0 p_1 p_2 ... (line contains all p_i that correspond to the model variable mentioned in the previous line)
		// keeps alternating between these two types of lines
		results = new ArrayList<Variable>();
		int nmbModelVariables = 0;
		try {
			BufferedReader in = new BufferedReader(new FileReader(variableFile));
			String line;
			String[] vars;
			Variable variable;
			String name;
			while ((line = in.readLine()) != null) {
				// model variable name
				name = line.trim();
				nmbModelVariables++;
				
				// parsing and converting the p_i
				line = in.readLine();
				vars = line.split(" ");

				// saving info related to the model information
				variable = new Variable(vars.length);
				variable.setName(name);

				for (int i = 0; i < vars.length; i++) {
					if (!vars[i].equals(""))
						variable.getSatVariables().add(Integer.parseInt(vars[i]) - 1);
				}

				results.add(variable);
			}
		} catch (IOException e) {
			System.err.println("Problems reading the variable file");
			System.exit(2);
		}

		ArrayList<Integer> variables = new ArrayList<Integer>(1000);
		for (int i = 0; i < results.size(); i++){
			variables.addAll(results.get(i).getSatVariables());
		}
		
		if (irrel) {
			BitSet r = Irrelevance.checkIrrelevance(proofFile, variables);
			
			int k = variables.size() / nmbModelVariables;
			for (int i = 0; i < results.size(); i++){
				results.get(i).setIrrel(r.get(i * k, (i + 1) * k));
			}

			// Now print out results
			PrintWriter out = new PrintWriter(new FileWriter(directory + "/output_irrel_" + suffix));
			for (int i = 0; i < results.size(); i++){
				out.println(results.get(i).getName() + "\t" + results.get(i).isIrrelevant() + "\t" + results.get(i).getIrrel());
			}
			out.close();
		}

		if (localIrrel) {
			BitSet r = LocalIrrelevance.checkLocalIrrelevance(proofFile, modelFile, propertyFile, variables);

			int k = variables.size() / nmbModelVariables;
			for (int i = 0; i < results.size(); i++){
				results.get(i).setLocalIrrel(r.get(i * k, (i + 1) * k));
			}

			// Now print out results
			PrintWriter out = new PrintWriter(new FileWriter(directory + "/output_local_irrel_" + suffix));
			for (int i = 0; i < results.size(); i++){
				out.println(results.get(i).getName() + "\t" + results.get(i).isLocalIrrelevant() + "\t" + results.get(i).getLocalIrrel());
			}
			out.close();
		}

		if (periph) {
			BitSet r = Peripherality.checkPeripherality(proofFile, modelFile, propertyFile, nmrP_i, variables, lastIndex);

			int k = variables.size() / nmbModelVariables;
			for (int i = 0; i < results.size(); i++){
				results.get(i).setPeriph(r.get(i * k, (i + 1) * k));
			}

			// Now print out results
			PrintWriter out = new PrintWriter(new FileWriter(directory + "/output_periph_" + suffix));
			for (int i = 0; i < results.size(); i++){
				out.println(results.get(i).getName() + "\t" + results.get(i).isPeripheral() + "\t" + results.get(i).getPeriph());
			}
			out.close();
		}

		


	}

	public static List<Variable> getResults() {
		return results;
	}

	public static void setResults(List<Variable> results) {
		VaqTree.results = results;
	}

}
