package edu.toronto.cs.algorithms;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import edu.toronto.cs.info.Clause;
import edu.toronto.cs.parsers.PeripheralityProofParser;
import edu.toronto.cs.util.Constants;
import edu.toronto.cs.util.StopWatch;

public class Peripherality {

	// proofs have to be preprocessed using the ProofReindexParser before
	// applying peripherality
	public static ArrayList<Clause> proofTree;
	public static ArrayList<Integer> clauseOrder;

	public static DefaultHandler handler;

	public static SAXParserFactory factory = SAXParserFactory.newInstance();

	public static BitSet checkPeripherality(File proofFile, File modelFile, File propertyFile, int nmrP_i, ArrayList<Integer> variables, int lastIndex)
			throws IOException {
		StopWatch stopWatch = new StopWatch();

		try {

			proofTree = new ArrayList<Clause>();
			clauseOrder = new ArrayList<Integer>();
			handler = new PeripheralityProofParser(lastIndex);

			SAXParser saxParser = factory.newSAXParser();

			stopWatch.reset();
			saxParser.parse(proofFile, handler);
			stopWatch.pause();

			System.out.println("Generate proof tree: " + stopWatch);

		} catch (SAXException ex) {
			System.err.println("SAXException " + ex.getMessage());
			System.exit(2);
		} catch (ParserConfigurationException ex) {
			System.err.println("ParserCopnfigurationException " + ex.getMessage());
			System.exit(2);
		} catch (IOException ex) {
			System.err.println("IOException " + ex.getMessage());
			System.exit(2);
		}

		// at this point, proofTree has been generated (already includes model
		// and property info for root clauses)

		BitSet results = new BitSet(variables.size());

		// apply peripherality (modifies results)
		int size = variables.size();
		int end;
		for (int i = 0; i < size; i += nmrP_i) {
			end = i + nmrP_i;
			if (end > size)
				end = size;

			stopWatch.reset();
			setPeripheralityLabels(variables.subList(i, end), i, results);
			stopWatch.pause();

			System.out.println("Peripherality " + i + " " + end + ": " + stopWatch);
		}

		return results;
	}

	private static void setPeripheralityLabels(List<Integer> testVariables, int to, BitSet results) {

		StopWatch stopWatch = new StopWatch();

		int sizeTest = testVariables.size();

		int size = proofTree.size();

		System.out.println("Number clauses in proof: " + size);
		
		// cleaning up from previous peripherality runs on this tree
		if (to > 0) {
			for (int i = 0; i < size; i++) {
				proofTree.get(i).processed = false;
			}
		}

		Clause clause;

		LinkedList<Clause> frontier = new LinkedList<Clause>();

		Clause emptyClause = new Clause();

		stopWatch.pause();
		System.out.println("setup " + stopWatch);
		stopWatch.reset();

		size = clauseOrder.size();
		int index;
		for (int i = 0; i < size; i++) {
			index = clauseOrder.get(i);
			
//			System.out.print(index + " ");
			clause = proofTree.get(index);

			if (clause.isRoot()) {

				clause.peripheraliyLabels = new BitSet(sizeTest * 2);
				clause.inBothLabel = new BitSet(sizeTest);

				for (int j = 0; j < sizeTest; j++) {
					int variable = testVariables.get(j);

					// p_i not in clause?
					if (!(clause.variables.contains(variable) || clause.variables.contains(-1 * variable))) {
						setPeripheralityLabel(clause, j, Constants.LABEL_TYPE_NONE);
					} else {
						// p_i in clause and:

						int clauseMP = clause.mpInfo;
						if (clauseMP == 1) {
							// clause only appears in CL_K
							setPeripheralityLabel(clause, j, Constants.LABEL_TYPE_MODEL);
						} else if (clauseMP == 2) {
							// only appears in CL_e
							setPeripheralityLabel(clause, j, Constants.LABEL_TYPE_PROPERTY);
						} else {
							// appears in both CL_K, CL_e
							clause.inBothLabel.set(j);
						}
					}
				}
				clause.processed = true;
			} else {
				if (clause.variables.size() == 0) {
					emptyClause = clause;
				}
				if (clause.leftParent.processed && clause.rightParent.processed) {
					processDerivedClause(testVariables, sizeTest, clause);
					
					clause.leftParent.unprocessedChildren--;
					if (clause.leftParent.unprocessedChildren == 0)
						Peripherality.proofTree.set(clause.leftParent.index, null);
					clause.leftParent = null;

					clause.rightParent.unprocessedChildren--;
					if (clause.rightParent.unprocessedChildren == 0)
						Peripherality.proofTree.set(clause.rightParent.index, null);
					clause.rightParent = null;
				} else {
					frontier.add(clause);
				}
			}
		}

		stopWatch.pause();
		System.out.println("initial pass through clauses " + stopWatch);
		stopWatch.reset();

		// go through remaining nodes
		while (frontier.size() > 0) {
			clause = frontier.removeFirst();

			if (!(clause.leftParent.processed && clause.rightParent.processed)) {
				frontier.add(clause);
			} else {
				processDerivedClause(testVariables, sizeTest, clause);

				clause.leftParent.unprocessedChildren--;
				if (clause.leftParent.unprocessedChildren == 0)
					Peripherality.proofTree.set(clause.leftParent.index, null);
				clause.leftParent = null;

				clause.rightParent.unprocessedChildren--;
				if (clause.rightParent.unprocessedChildren == 0)
					Peripherality.proofTree.set(clause.rightParent.index, null);
				clause.rightParent = null;
			}

		}
		stopWatch.pause();
		System.out.println("loop through frontier " + stopWatch);
		stopWatch.reset();

		// condition for peripherality of z:
		// (emptyClause.inBothLabel.get(z) || getPeripheralityLabel(emptyClause,
		// z) != Constants.LABEL_TYPE_MODEL_AND_PROPERTY)
		for (int z = 0; z < sizeTest; z++) {
			if (!(emptyClause.inBothLabel.get(z) || getPeripheralityLabel(emptyClause, z) != Constants.LABEL_TYPE_MODEL_AND_PROPERTY)) {
				results.set(z + to);
				// System.out.print((z + to) + " ");
			}
		}

		stopWatch.pause();
		System.out.println("saving results " + stopWatch);
		stopWatch.reset();
	}

	private static void processDerivedClause(List<Integer> testVariables, int sizeTest, Clause clause) {

		int resolIndex;
		boolean leftInBoth;
		boolean rightInBoth;
		int label;
		int leftLabel;
		int rightLabel;
		clause.peripheraliyLabels = new BitSet(sizeTest * 2);
		clause.inBothLabel = new BitSet(sizeTest);

		resolIndex = testVariables.indexOf(clause.reductionVariable);

		if (resolIndex > -1) {
			// handle special case for resolution variable
			leftInBoth = clause.leftParent.inBothLabel.get(resolIndex);
			rightInBoth = clause.rightParent.inBothLabel.get(resolIndex);

			rightLabel = getPeripheralityLabel(clause.rightParent, resolIndex);
			leftLabel = getPeripheralityLabel(clause.leftParent, resolIndex);

			if (!(leftInBoth && rightInBoth))
				if (rightLabel != leftLabel)
					setPeripheralityLabel(clause, resolIndex, Constants.LABEL_TYPE_MODEL_AND_PROPERTY);
				else if (leftInBoth)
					if (rightLabel == Constants.LABEL_TYPE_NONE || rightLabel == Constants.LABEL_TYPE_MODEL_AND_PROPERTY)
						setPeripheralityLabel(clause, resolIndex, Constants.LABEL_TYPE_MODEL_AND_PROPERTY);
					else if (rightInBoth)
						if (leftLabel == Constants.LABEL_TYPE_NONE || leftLabel == Constants.LABEL_TYPE_MODEL_AND_PROPERTY)
							setPeripheralityLabel(clause, resolIndex, Constants.LABEL_TYPE_MODEL_AND_PROPERTY);

			// handle the rest of the variables
			for (int z = 0; z < sizeTest; z++) {
				if (z == resolIndex)
					continue;

				leftInBoth = clause.leftParent.inBothLabel.get(z);
				rightInBoth = clause.rightParent.inBothLabel.get(z);

				label = getPeripheralityLabel(clause.leftParent, z);
				leftLabel = getPeripheralityLabel(clause.leftParent, z);
				rightLabel = getPeripheralityLabel(clause.rightParent, z);

				if (leftInBoth && rightInBoth) { // see 4
					clause.inBothLabel.set(z);
				} else if (leftInBoth || rightInBoth) { // see 5
					if (leftInBoth) {
						label = rightLabel;
					}
					if (label == Constants.LABEL_TYPE_NONE) {
						clause.inBothLabel.set(z);
					} else {
						setPeripheralityLabel(clause, z, label);
					}
				} else {
					// see 6
					setPeripheralityLabel(clause, z, leftLabel | rightLabel);
				}
			}

		} else {
			// handle all variables in the same way
			for (int z = 0; z < sizeTest; z++) {

				leftInBoth = clause.leftParent.inBothLabel.get(z);
				rightInBoth = clause.rightParent.inBothLabel.get(z);

				label = getPeripheralityLabel(clause.leftParent, z);
				leftLabel = getPeripheralityLabel(clause.leftParent, z);
				rightLabel = getPeripheralityLabel(clause.rightParent, z);

				if (leftInBoth && rightInBoth) { // see 4
					clause.inBothLabel.set(z);
				} else if (leftInBoth || rightInBoth) { // see 5
					if (leftInBoth) {
						label = rightLabel;
					}
					if (label == Constants.LABEL_TYPE_NONE) {
						clause.inBothLabel.set(z);
					} else {
						setPeripheralityLabel(clause, z, label);
					}
				} else {
					// see 6
					setPeripheralityLabel(clause, z, leftLabel | rightLabel);
				}
			}

		}
		clause.processed = true;
	}

	public static int getPeripheralityLabel(Clause clause, Integer index) {

		boolean f1 = clause.peripheraliyLabels.get(2 * index);
		boolean f2 = clause.peripheraliyLabels.get(2 * index + 1);

		if (f1 & f2)
			return Constants.LABEL_TYPE_MODEL_AND_PROPERTY;
		else if (!f1 & f2)
			return Constants.LABEL_TYPE_MODEL;
		else if (f1 & !f2)
			return Constants.LABEL_TYPE_PROPERTY;
		else
			return Constants.LABEL_TYPE_NONE;
	}

	public static void setPeripheralityLabel(Clause clause, Integer index, int label) {
		assert ((label & 3) == label) : "Label out of range: " + label;

		switch (label) {
		case Constants.LABEL_TYPE_NONE:
			clause.peripheraliyLabels.set(2 * index, false);
			clause.peripheraliyLabels.set(2 * index + 1, false);
			break;
		case Constants.LABEL_TYPE_MODEL:
			clause.peripheraliyLabels.set(2 * index + 1);
			break;
		case Constants.LABEL_TYPE_PROPERTY:
			clause.peripheraliyLabels.set(2 * index);
			break;
		case Constants.LABEL_TYPE_MODEL_AND_PROPERTY:
			clause.peripheraliyLabels.set(2 * index);
			clause.peripheraliyLabels.set(2 * index + 1);
			break;
		}
	}

}
